# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import ast
import json
import os
import shutil
import string
from string import Template

from git import Repo

from orchd_sdk import util
from orchd_sdk.models import Project


PROJECT_TEMPLATE_REPO = 'https://github.com/iiot-orchestrator/orchd-project-template.git'
PROJECT_TEMPLATE_DIR = '.orchd'
REACTION_TEMPLATE_FILE = os.path.join(PROJECT_TEMPLATE_DIR,
                                      'src/pkg_name/reactions/base_reaction.py.template')
SENSOR_TEMPLATE_FILE = os.path.join(PROJECT_TEMPLATE_DIR,
                                    'src/pkg_name/sensors/base_sensor.py.template')
SINK_TEMPLATE_FILE = os.path.join(PROJECT_TEMPLATE_DIR,
                                  'src/pkg_name/sinks/base_sink.py.template')


class ProjectBootstrapper:

    @staticmethod
    def setup_project(project: Project):
        folder_handler = ProjectFoldersHandler(project)
        folder_handler.prepare_project_folders()
        ProjectTemplateFilesHandler(project).process_templates()


class ProjectFoldersHandler:

    def __init__(self, project: Project):
        self.project = project

    def prepare_project_folders(self):
        self.clone_orchd_project_template()
        self.copy_skeleton()
        self.rename_main_package_folder()

    def copy_skeleton(self):
        shutil.copytree(self.template_dir(), self.project_dir(), dirs_exist_ok=True,
                        ignore=lambda src, names: [n for n in names if n in self.initial_ignored_files()])

    @staticmethod
    def initial_ignored_files():
        return ['base_sink.py.template', 'base_sensor.py.template', 'base_reaction.py.template']

    def rename_main_package_folder(self):
        shutil.move(f'{self.project_dir()}/src/pkg_name', f'{self.project_dir()}/src/{self.project.main_package}')

    @staticmethod
    def is_template_folder_present():
        return os.path.exists(PROJECT_TEMPLATE_DIR)

    def clone_orchd_project_template(self):
        Repo.clone_from(PROJECT_TEMPLATE_REPO, self.template_dir())
        shutil.rmtree(os.path.join(self.template_dir(), '.git'))

    def template_dir(self):
        return os.path.join(self.project_dir(), PROJECT_TEMPLATE_DIR)

    def project_dir(self):
        return os.path.join(os.getcwd(), self.project.name)


class ProjectTemplateFilesHandler:

    def __init__(self, project):
        self.project = project

    def process_templates(self):
        self.create_project_meta_file()
        self.create_version_file()
        self.create_setup_cfg_file()

    def create_project_meta_file(self):
        with open(f'{self.project.name}/orchd.meta.json', 'w') as fd:
            fd.write(json.dumps(self.project.dict(), indent=2) + '\n\n')

    def create_version_file(self):
        version_file_template = self.load_version_file_template()
        with open(f'{self.project.name}/VERSION', 'w') as version_file:
            version_file_content = string.Template(version_file_template).substitute(
                version=self.project.version
            )
            version_file.write(version_file_content + '\n\n')

    def load_version_file_template(self):
        with open(f'{self.project.name}/VERSION', 'r') as version_file:
            return version_file.read()

    def create_setup_cfg_file(self):
        setup_cfg_template = self.load_setup_cfg_template()
        with open(f'{self.project.name}/setup.cfg', 'w') as setup_cfg:
            setup_cfg_content = string.Template(setup_cfg_template).substitute(
                name=self.project.name,
                description=self.project.description,
                license=self.project.license
            )
            setup_cfg.write(setup_cfg_content + '\n\n')

    def load_setup_cfg_template(self):
        with open(f'{self.project.name}/setup.cfg', 'r') as setup_cfg_file:
            return setup_cfg_file.read()


class ReactionProperties:

    def __init__(self, project, reaction_module_name, version, triggers, sinks, handler_params, active):
        self.project = project
        self.reaction_module_name = reaction_module_name
        self.version = version
        self.triggers = triggers
        self.sinks = sinks
        self.handler_params = handler_params
        self.active = active

    def reaction_module_fq_name(self):
        return f'{self.project.main_package}.reactions.{self.reaction_module_name}'

    def reaction_class_name(self):
        return f'{util.snake_to_camel_case(self.reaction_module_name)}Reaction'

    def reaction_class_fq_name(self):
        return f'{self.reaction_module_fq_name()}.{self.reaction_class_name()}'

    def reaction_handler_class_name(self):
        return f'{self.reaction_class_name()}Handler'

    def reaction_handler_class_fq_name(self):
        return f'{self.reaction_module_fq_name()}.{self.reaction_handler_class_name()}'

    def reaction_namespaced_name(self):
        return f'{self.project.namespace}.{self.reaction_class_name()}'

    @staticmethod
    def triggers_list_from_str(triggers_str):
        try:
            triggers_as_list = ast.literal_eval(triggers_str)
            if type(triggers_as_list) is not list:
                raise ValueError(
                    'Triggers not seems to be a valid list. Use the format ["trigger1", "trigger2", "etc..."]')
            return triggers_as_list
        except ValueError:
            raise ValueError('Not able to process list of triggers. Verify its value.')


class ReactionTemplateFileHandler:

    def __init__(self, project: Project, reaction_properties: ReactionProperties):
        self.project = project
        self.reaction_properties = reaction_properties

    def create_reaction_module_file(self):
        self.copy_reaction_template_module_file()
        self.process_reaction_template()

    def copy_reaction_template_module_file(self):
        shutil.copyfile(REACTION_TEMPLATE_FILE, self.reaction_module_file_path())

    def process_reaction_template(self):
        reaction_template_str = Template(self.load_reaction_template_file())
        reaction_content = reaction_template_str.substitute(
            reaction_name=self.reaction_properties.reaction_namespaced_name(),
            reaction_class_name=self.reaction_properties.reaction_class_name(),
            reaction_handler_class_name=self.reaction_properties.reaction_handler_class_name(),
            reaction_handler_fqn_class=self.reaction_properties.reaction_handler_class_fq_name(),
            version=self.reaction_properties.version,
            triggers=self.reaction_properties.triggers,
            handler_params=self.reaction_properties.handler_params,
            active=self.reaction_properties.active,
            sinks=self.reaction_properties.sinks
        )
        self.save_reaction_file(reaction_content)

    def reaction_module_file_path(self):
        return f'./src/{self.project.main_package}/reactions/{self.reaction_properties.reaction_module_name}.py'

    def load_reaction_template_file(self):
        with open(self.reaction_module_file_path(), 'r') as reaction_template_file:
            return reaction_template_file.read()

    def save_reaction_file(self, content):
        with open(self.reaction_module_file_path(), 'w') as reaction_file:
            reaction_file.write(content)


class ReactionBootstrapper:

    @staticmethod
    def create(project: Project, reaction_properties: ReactionProperties):
        template_file_handler = ReactionTemplateFileHandler(project, reaction_properties)
        template_file_handler.create_reaction_module_file()


class SensorProperties:
    def __init__(self, project, sensor_module_name, description, version, sensor_param,
                 sensing_interval, communicator):
        self.project = project
        self.sensor_module_name = sensor_module_name
        self.description = description
        self.version = version
        self.sensor_param = sensor_param
        self.sensing_interval = sensing_interval
        self.communicator = communicator

    def sensor_namespaced_name(self):
        return f'{self.project.namespace}.{self.sensor_class_name()}'

    def sensor_module_fq_name(self):
        return f'{self.project.main_package}.sensors.{self.sensor_module_name}'

    def sensor_class_name(self):
        return util.snake_to_camel_case(self.sensor_module_name) + 'Sensor'

    def sensor_class_fq_name(self):
        return f'{self.sensor_module_fq_name()}.{self.sensor_class_name()}'


class SensorTemplateFileHandler:

    def __init__(self, project: Project, sensor_properties: SensorProperties):
        self.project = project
        self.sensor_properties = sensor_properties

    def create_sensor_module_file(self):
        self.copy_sensor_module_template_file()
        self.process_sensor_template()

    def copy_sensor_module_template_file(self):
        shutil.copy(SENSOR_TEMPLATE_FILE, self.sensor_module_file_path())

    def sensor_module_file_path(self):
        return f'./src/{self.project.main_package}/sensors/{self.sensor_properties.sensor_module_name}.py'

    def process_sensor_template(self):
        sensor_module_template = Template(self.load_sensor_template_file())
        content = sensor_module_template.substitute(
            sensor_name=self.sensor_properties.sensor_namespaced_name(),
            sensor_class_name=self.sensor_properties.sensor_class_name(),
            sensor_class_fqn=self.sensor_properties.sensor_class_fq_name(),
            communicator_class_fqn=self.sensor_properties.communicator,
            version=self.sensor_properties.version,
            description=self.sensor_properties.description,
            parameters=self.sensor_properties.sensor_param,
            sensing_interval=self.sensor_properties.sensing_interval
        )
        self.save_sensor_file(content)

    def load_sensor_template_file(self):
        with open(self.sensor_module_file_path()) as sensor_template_str:
            return sensor_template_str.read()

    def save_sensor_file(self, content):
        with open(self.sensor_module_file_path(), 'w') as sensor_file:
            sensor_file.write(content + '\n\n')


class SensorBootstrapper:

    @staticmethod
    def create(project: Project, sensor_properties: SensorProperties):
        sensor_template_handler = SensorTemplateFileHandler(project, sensor_properties)
        sensor_template_handler.create_sensor_module_file()


class SinkProperties:

    def __init__(self, project: Project, sink_module_name: str, version: str, parameters: str):
        self.project = project
        self.sink_module_name = sink_module_name
        self.version = version
        self.parameters = parameters

    def sink_class_name(self):
        return util.snake_to_camel_case(self.sink_module_name) + 'Sink'

    def sink_module_fq_name(self):
        return f'{self.project.main_package}.sinks.{self.sink_module_name}'

    def sink_class_fq_name(self):
        return f'{self.sink_module_fq_name()}.{self.sink_class_name()}'

    def sink_namespaced_name(self):
        return f'{self.project.namespace}.{self.sink_module_name}.{self.sink_class_name()}'

    def sink_json_parameters_as_dict(self):
        return json.loads(self.parameters)


class SinkTemplateFileHandler:

    def __init__(self, project: Project, sink_properties: SinkProperties):
        self.project = project
        self.sink_properties = sink_properties

    def create_sink_module_file(self):
        self.copy_sink_template_file()
        self.process_sink_template_file()

    def process_sink_template_file(self):
        template = Template(self.load_sink_module_file())
        content = template.substitute(
            sink_class_name=self.sink_properties.sink_class_name(),
            sink_class_fq_name=self.sink_properties.sink_class_fq_name(),
            sink_namespaced_name=self.sink_properties.sink_namespaced_name(),
            sink_version=self.sink_properties.version,
            sink_properties=self.sink_properties.parameters
        )
        self.save_sink_file(content)

    def save_sink_file(self, content):
        with open(self.sink_module_file_path(), 'w') as sink_module_file:
            sink_module_file.write(content + '\n\n')

    def copy_sink_template_file(self):
        shutil.copy(SINK_TEMPLATE_FILE, self.sink_module_file_path())

    def sink_module_file_path(self):
        return f'./src/{self.project.main_package}/sinks/{self.sink_properties.sink_module_name}.py'

    def load_sink_module_file(self):
        with open(self.sink_module_file_path(), 'r') as sink_module_file:
            return sink_module_file.read()


class SinkBootstrapper:

    @staticmethod
    def create(project: Project, sink_properties: SinkProperties):
        handler = SinkTemplateFileHandler(project, sink_properties)
        handler.create_sink_module_file()



# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import uuid

from typing import Dict, List, Any, Union, Optional

from pydantic import Field, BaseModel

import orchd_sdk
from orchd_sdk import util


def uuid_str_factory():
    return str(uuid.uuid4())


class Ref(BaseModel):
    """
    Represents an Id and associated possible additional info.

    This model is intended to be used when sending Id over the network.
    Many operations, e.g. HTTP services requires IDs to be sent as payload,
    this model makes it easy to serialized/deserialize the Id in the source
    and destination.
    """

    id: str = Field(
        title='ID',
        description='An ID.',
        example='0ccb8b84-c52d-4a6a-af6b-a2c273745825'
    )

    metadata: Any = Field(
        title='ID Metadata',
        description='Optional field with Key/Value pairs with additional info about '
                    'the id.',
        example={'related_model': 'SomeModelClass'}
    )


class Event(BaseModel):
    """
    Representation of an event of the Orchd system.

    An event have a name and data associated. Each event have
    an unique ID. They are propagated into the Orchd System usually
    by Sensors and captured by Reactions that can react to it.
    """
    event_name: str = Field(
        title='Event Name',
        description='A unique/namespaced name for the event.',
        regex=r'^\w[\w\._\-]+$',
        example='io.orchd.events.system.Test'
    )
    data: Dict[str, Any] = Field(
        default_factory=dict,
        title='Event Data',
        description='Data attached to the event in the form of key/value pairs.'
    )
    id: str = Field(
        default_factory=uuid_str_factory,
        title='Event ID',
        description='Event Unique Identifier.'
    )

    class Config:
        additional_properties = False


class SinkTemplate(BaseModel):
    """
    Representation of an Orchd Sink

    The properties are Sink implementation specific and the allowed options must be
    defined in the Sink implementation docs.
    """
    id: str = Field(
        default_factory=uuid_str_factory,
        title='Sink Template Id',
        description='Id of the Sink Template',
        example='3d2077c7-dbad-4975-b769-a8da870cf5f6'
    )

    name: str = Field(
        title='Sink\'s Template name',
        description='A descriptive name for the Sink.',
        example='io.orchd.sinks.MyMainSink'
    )

    version: str = Field(
        title='Sink\'s Template Version',
        description='Version of this Sink.',
        example='0,1'
    )

    sink_class: str = Field(
        title='Sink Template Class',
        description='Sink Class to be used to instantiate sinks.',
        example='orchd_sdk.sink.DummySink'
    )

    properties: Dict[str, str] = Field(
        title='Sink Properties',
        description='key/value pairs with Sink specific properties',
        example={'endpoint': 'http//www.example.com/test'}
    )


class Sink(BaseModel):
    """
    Sink instance information.

    Sink instances are attached to Reactions, this model represents the status of
    sink instances.
    """
    id: str = Field(
        default_factory=uuid_str_factory,
        title='Sink Id',
        description='Id of the Sink',
        example='3d2077c7-dbad-4975-b769-a8da870cf5f6'
    )

    template: SinkTemplate = Field(
        title='Sink Template',
        description='Sink Template used to instantiate the Sink',
        example=SinkTemplate(sink_class='orchd_sdk.sink.DummySink',
                             name='io.orchd.sinks.DummySink',
                             version='0.1',
                             properties={'endpoint': 'https://example.com/test'})
    )


class ReactionTemplate(BaseModel):
    """
    Representation of a Reaction Template to be used to create `Reaction`.

    Nodes can react on events detected in the network or internally in the node.
    Some example of possible events are:
    - Service Discovered
    - USB Device Attached
    - Service Down
    """
    name: str = Field(
        title='Reaction Template Name',
        description='A unique/namespaced name for the Reaction Template',
        regex=r'^\w[\w\._\-]+$',
        example='io.orchd.reaction_template.DummyTemplate'
    )
    version: str = Field(
        title='Template Version',
        description='Version of this reaction template.',
        example='1.0'
    )
    handler: str = Field(
        title='Handler Class',
        description='The full name/path of the handler Class',
        example='orchd_sdk.reaction.DummyReactionHandler'
    )
    triggered_on: List[str] = Field(
        title='Triggered On',
        description='List of event names that triggers the handler.',
        example=['io.orchd.events.system.Test']
    )
    handler_parameters: Dict[str, str] = Field(
        default_factory=dict,
        title='Handler Parameters',
        description='Parameters to be passed to the handler in the form key/value '
                    'pair.',
        example={'test_type': 'full'},
    )

    sinks: Optional[List[SinkTemplate]] = Field(
        title='Sinks',
        default_factory=list,
        description='Sinks used by reactions created from this template.',
        example=[SinkTemplate(sink_class='orchd_sdk.sink.DummySink',
                              name='io.orchd.sinks.DummySink',
                              version='0.1',
                              properties={'endpoint': 'https://example.com/test'})]
    )

    active: bool = Field(
        default=True,
        title='Active',
        description='Activate/Deactivate the availability of the template to create '
                    'new Reactions.',
        example=True
    )
    id: str = Field(
        default_factory=uuid_str_factory,
        title='Reaction Template ID',
        description='Unique Reaction Template identifier.',
        example='0a0866da-2a41-41a3-bcd9-9be9eedb2525'
    )


class SensorTemplate(BaseModel):
    """
    Representation of a Sensor Template describing the Sensor characteristics.

    A Sensor is intended to detect events outside the system and inject tem into
    it. The SensorTemplate is responsible to hold metadata about a Sensor to be
    instantiated. By having a SensorTemplate it is possible to instantiate a Sensor.
    Important information comes with the template, as an example a set of properties
    that can be used by the sensor to configure its behavior.
    """
    id: str = Field(
        title='Id',
        description='Unique identification of the Sensor',
        example='0ba3376f-64b8-4ecf-a579-66c353100e1c',
        default_factory=uuid_str_factory
    )

    name: str = Field(
        title='Sensor\'s Name',
        description='The namespaced name of the Sensor',
        regex=r'^\w[\w\._\-]+$',
        example='io.orchd.sensor.template.DummySensorTemplate'
    )

    version: str = Field(
        title='Version',
        description='Sensor\'s version',
        example='1.0'
    )

    sensor: str = Field(
        title='Sensor Class',
        description='Class that implements the Sensor',
        example='orchd_sdk.sensor.DummySensor'
    )

    sensing_interval: float = Field(
        title='Sensing Interval',
        description='Interval between two consecutive sense calls in seconds.',
        example=0.1
    )

    communicator: str = Field(
        title='Communicator Class',
        description='Class of the Communicator to used.',
        example='orchd_sdk.sensor.LocalCommunicator'
    )

    parameters: Dict[str, Union[str, int, float, List[Union[str, int, float]]]] = \
        Field(title='Sensor Parameters',
              description='Parameters to be used by the Sensor during Runtime',
              example={'poll_interval': 3}
    )

    description: str = Field(
        title='Description',
        description='Description of the Sensor',
        example='Sense for changes in a Dummy value in the System'
    )


class ReactionInfo(BaseModel):
    """
    Represents the state and info of an Reaction instance.
    """
    id: str = Field(
        title='Reaction Id',
        description='Id of the Reaction.',
        example='cdc51818-6ebe-4931-bed2-2a297c70681e'
    )

    state: str = Field(
        title='State',
        description='Current Reaction state.',
        example='READY'
    )

    template: ReactionTemplate = Field(
        title='Reaction Template',
        description='Reaction Template used to instantiate this Reaction.',
        example='d818d8ff-f859-4239-8602-103563b8a2ff'
    )

    sinks_instances: List[Sink] = Field(
        title='Sinks\' IDs',
        description='Id of Sinks associated to this reaction',
        example=['1d1c79de-7170-4dc6-aae1-9504ea7c793e',
                 'eaaf2065-484e-4c31-88fa-c3b422aa8927']
    )


class Sensor(BaseModel):
    """
    Represents the state and data of a Sensor
    """
    id: str = Field(
        title='Sensor Id',
        description='Id of the sensor.',
        example='7447f5f8-63f6-48d0-8537-5fae0b30015d'
    )

    template: SensorTemplate = Field(
        title='Sensor Template',
        description='Template related to the Sensor.',
        example='orchd_sdk.sensor.DummySensor'
    )

    status: tuple = Field(
        title='Sensor Status',
        description='Current status of the Sensor.',
        example=(2, 'READY')
    )

    events_count: Optional[int] = Field(
        title='Events Counter',
        description='Number of events sensed by the sensor.',
        example=1002
    )

    events_forwarded: Optional[int] = Field(
        title='Forwarded Events',
        description='Number of events sensed and forwarded to Orchd.',
        example=540
    )

    events_discarded: Optional[int] = Field(
        title='Events discarded',
        description='Number of events sensed, captured but discarded.'
    )


class Project(BaseModel):
    class Config:
        validate_assignment = True

    def __init__(self, **kwargs):
        main_package_name = util.kebab_case_to_snake_case(kwargs.get('name'))
        kwargs['main_package'] = main_package_name
        super().__init__(**kwargs)

    orchd_sdk_version = orchd_sdk.version()
    version: str = '0.0'
    author: str = 'unknown'
    description: str = 'no description'
    license: str = 'unknown'
    name: str = Field(
        title='Name of the Project',
        pattern='^((_?[a-z0-9])*|(_?[A-Z0-9])*|_)$',
        default='my_orchd_project'
    )
    namespace: str = ''
    main_package: str = 'project'
    reactions: Dict[str, str] = {}
    sensors: Dict[str, str] = {}
    sinks: Dict[str, str] = {}

    def add_reaction(self, name: str, handler: str):
        self.reactions[name] = handler

    def add_sensor(self, name: str, class_: str):
        self.sensors[name] = class_

    def add_sink(self, name: str, class_: str):
        self.sinks[name] = class_




# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio
import importlib
import logging
import sys
import uuid

from abc import abstractmethod, ABC
from asyncio import AbstractEventLoop
from typing import Any, Dict, List, Union, Tuple

from reactivex.observer import Observer
from reactivex.subject import Subject

from orchd_sdk.common import import_class
from orchd_sdk.errors import SinkError, ReactionHandlerError, ReactionError
from orchd_sdk.models import Event, ReactionTemplate, SinkTemplate, ReactionInfo
from orchd_sdk.sink import AbstractSink, DummySink

logger = logging.getLogger(__name__)


class ReactionsEventBus:
    """
    The Reaction Event Bus.

    The Reaction Event Bus wraps a reactivex.subject.Subject and offers
    a method to register Reactions (they are reactivex.observer.Observer).

    Whenever ones wants to propagate an event on the system CAN do this
    through an global reaction event bus. However it is allowed to create
    more BUSES depending on the system architecture being implemented.
    """
    def __init__(self):
        self._subject = Subject()

    def register_reaction(self, reaction: "Reaction"):
        """Registers a Reaction (Observer) on the subject."""
        disposable = self._subject.subscribe(reaction)
        reaction.disposable = disposable

    def event(self, event_: Event):
        """Forwards the event to the subscribers"""
        self._subject.on_next(event_)

    def remove_all_reactions(self):
        """Unsubscribe all observers"""
        NotImplementedError()


class ReactionHandler(ABC):
    """
    A Reaction handler for a event.
    """

    @abstractmethod
    def handle(self, event: Event, reaction: ReactionTemplate) -> None:
        """
        Code to be executed as an reaction to an event.

        :param event: The event that triggered the action.
        :param reaction: The reaction object.
        """


class ReactionState:
    UNINITIALIZED = (1, 'PROVISIONING')
    READY = (2, 'READY')
    RUNNING = (3, 'RUNNING')
    STOPPED = (4, 'STOPPED')
    ERROR = (5, 'ERROR')
    FINALIZED = (6, 'FINALIZED')


class ReactionSinkManager:

    def __init__(self, reaction):
        self._sinks: Dict[str, AbstractSink] = dict()
        self.reaction: Reaction = reaction

    @property
    def sinks(self) -> List[AbstractSink]:
        return list(self._sinks.values())

    def add_sink(self, sink_template: SinkTemplate):
        try:
            SinkClass = import_class(sink_template.sink_class)
            sink: AbstractSink = SinkClass(sink_template)
            self._sinks[sink.id] = sink
            return sink
        except ModuleNotFoundError as e:
            raise SinkError(f'Not able to load Sink class {sink_template.sink_class}. '
                            f'Is it in PYTHONPATH?') from e

    async def create_sinks(self, sink_templates: List[SinkTemplate]) -> Dict[str, AbstractSink]:
        for template in sink_templates:
            try:
                self.add_sink(template)
            except SinkError as e:
                for sink in self.sinks:
                    await sink.close()
                    del self._sinks[sink.id]
                raise e

        return self._sinks

    async def remove_sink(self, sink_id):
        try:
            sink = self._sinks[sink_id]
            await sink.close()
            del self._sinks[sink_id]
        except KeyError as e:
            raise SinkError(f'Sink with given ID{sink_id} not Found!') from e

    def get_sink_by_id(self, sink_id):
        try:
            return self._sinks[sink_id]
        except KeyError as e:
            raise SinkError(f'Sink with given ID({sink_id}) Not Found!') from e

    async def close(self):
        for sink in self._sinks.values():
            await sink.close()
        self._sinks = dict()


class Reaction(Observer):
    """
    Reaction handling management class.

    This class instantiates the reaction handler and subscribes the Reaction
    to the events that triggers it.
    """

    def __init__(self, reaction_template: ReactionTemplate):
        super().__init__()
        self.state: Tuple = ReactionState.UNINITIALIZED
        self.handler: Union[ReactionHandler, None] = None
        self.disposable: Union[ReactionHandler, None] = None
        self.id = str(uuid.uuid4())
        self.reaction_template: ReactionTemplate = reaction_template
        self._loop: AbstractEventLoop = asyncio.get_event_loop()
        self.sink_manager = ReactionSinkManager(self)

    async def init(self):
        try:
            self.handler = self.create_handler_object()
            await self.sink_manager.create_sinks(self.reaction_template.sinks)
        except SinkError as e:
            self.state = ReactionState.ERROR
            raise ReactionError("While creating reaction, an error occurred preparing Sinks.") from e
        except ReactionHandlerError as e:
            self.state = ReactionState.ERROR
            raise ReactionError("While creating reaction, an error occurred preparing Reaction Handlers.") from e

        self.state = ReactionState.READY
        return self

    @property
    def sinks(self) -> List[AbstractSink]:
        return list(self.sink_manager.sinks)

    def status(self):
        return ReactionInfo(
            id=self.id,
            state=self.state[1],
            template=self.reaction_template,
            sinks_instances=[s.info for s in self.sink_manager.sinks]
        )

    def create_handler_object(self) -> ReactionHandler:
        """Instantiate a :class:`ReactionHandler` indicated
        in the reaction template."""
        class_parts = self.reaction_template.handler.split('.')
        class_name = class_parts.pop()
        module_name = '.'.join(class_parts)

        try:
            if module_name not in sys.modules:
                importlib.import_module(module_name)
            HandlerClass = getattr(sys.modules.get(module_name), class_name)
            self.handler = HandlerClass()
            return self.handler
        except (ModuleNotFoundError, AttributeError) as e:
            raise ReactionHandlerError(f'Reaction Handler module/class '
                                       f'{self.reaction_template.handler} not found!') from e

    def on_next(self, event: Event) -> None:
        if event.event_name in self.reaction_template.triggered_on or \
                '' in self.reaction_template.triggered_on:
            self.sink(self.handler.handle(event, self.reaction_template))

    def sink(self, data):
        for sink in self.sink_manager.sinks:
            logger.info(f"Sink {sink.id} scheduled to be executed.")
            self._loop.create_task(sink.sink(data))

    def activate(self, event_bus: ReactionsEventBus):
        event_bus.register_reaction(self)
        logger.debug(f'Reaction for template {self.reaction_template.id} '
                     f'Activated. ID({self.id}).')
        self.state = ReactionState.RUNNING

    def dispose(self) -> None:
        self.disposable.dispose()
        super().dispose()
        self.disposable = None
        self.state = ReactionState.STOPPED

    async def stop(self):
        self.dispose()

    async def close(self):
        if self.state == ReactionState.RUNNING:
            self.dispose()
        await self.sink_manager.close()
        self.state = ReactionState.FINALIZED


class DummyReaction(Reaction):
    """Dummy Reaction used system tests during runtime."""

    template = ReactionTemplate(
        id='cfe5b2cd-fb15-4ca6-888f-6a770d1a4e6a',
        name='io.orchd.reaction_template.DummyTemplate',
        version='1.0',
        triggered_on=['io.orchd.events.system.Test'],
        handler="orchd_sdk.reaction.DummyReactionHandler",
        sinks=[DummySink.template],
        handler_parameters=dict(),
        active=True
    )

    def __init__(self, custom_template: ReactionTemplate = None):
        super().__init__(custom_template or DummyReaction.template)


class DummyReactionHandler(ReactionHandler):
    """
    A Dummy ReactionHandler that is used in system tests.
    """
    def handle(self, event: Event, reaction: ReactionTemplate) -> Any:
        logger.info(f'DummyReactionHandler.handle Called')


global_reactions_event_bus = ReactionsEventBus()
"""System wide ReactionEventBus"""



# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import uuid
import logging

from typing import Any
from abc import abstractmethod, ABC

from orchd_sdk.errors import SinkError
from orchd_sdk.common import import_class
from orchd_sdk.models import SinkTemplate, Sink


logger = logging.getLogger(__name__)


class AbstractSink(ABC):
    """
    Implements the logic for a Sink.

    Sinks have only one objective, based o the given information via the template
    tries to "sink" data somewhere. This SOMEWHERE is defined by the Sink developer.
    Sinks are ideal for forwarding or storing data processed by `Reactions`.

    Inform the Reaction the Sinks that it must use and the Reaction will call the `sink`
    method os the Sink automatically after finishing its job.

    An example would be a Orchd Sensor subscribed to a MQTT subject capturing the data,
    a Reaction could capture the data and process it sinking the data in a CoAPSink that
    forwards the data to a CoAP based system.
    """

    def __init__(self, template: SinkTemplate):
        self.id = str(uuid.uuid4())
        self._template = template
        self._info = Sink(id=self.id, template=template)

    @abstractmethod
    def close(self):
        pass

    @property
    def info(self):
        return self._info

    @abstractmethod
    async def sink(self, data: Any):
        pass


class DummySink(AbstractSink):
    """Dummy Sink for testing purposes"""

    template = SinkTemplate(id='fa8d0c8d-d525-4984-a091-28dbef21febe',
                            sink_class='orchd_sdk.sink.DummySink',
                            name='io.orchd.sinks.DummySink',
                            version='0.1',
                            properties={'endpoint': 'https://example.com/test'})

    def __init__(self, template: SinkTemplate = None):
        super().__init__(template or self.template)

    async def sink(self, data):
        logger.info(f'Data SUNK by Dummy! Actually, I did Nothing! :P {data}')

    async def close(self):
        pass


def sink_factory(template: SinkTemplate):
    try:
        Class = import_class(template.sink_class)
        if isinstance(Class, AbstractSink):
            raise SinkError('Sink class is not Valid!')
        return Class(template)
    except ModuleNotFoundError:
        raise SinkError('Sink class not found in PYTHONPATH.')




# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

class InvalidInputError(Exception):
    """
    Invalid input given.
    """


class ReactorError(Exception):
    """Raise on internal Errors in the Reactor."""


class SinkError(Exception):
    """ Raised on Sink errors."""


class ReactionError(Exception):
    """ Raised on Reaction Management and Operation errors."""


class ReactionHandlerError(Exception):
    """ Raised by ReactionHandler implementations."""



# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from setuptools import setup, find_packages

name = 'orchd-sdk'
version = open('src/orchd_sdk/VERSION').read().strip()
author = "Mathias Santos de Brito"

requirements = [
    'reactivex',
    'pydantic',
    'GitPython',
    'colorama'
]

test_requirements = [
    *requirements,
    'pytest',
    'pytest-asyncio',
    'coverage',
    'black',
    'mypy'
]

doc_requirements = [
    'sphinx',
    'sphinx_rtd_theme'
]

setup(
    name=name,
    version=version,
    description='SDK for Orchd Ecosystem Applications',
    keywords='service resource orchestration edge cloud',
    url='http://orchd.io',
    classifiers=[
        'License :: Other/Proprietary License',
        'Intended Audience :: Information Technology',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: System :: Distributed Computing'
    ],
    author=author,
    author_email='mathias.brito@me.com',

    install_requires=requirements,
    extras_require={
        'test': test_requirements,
        'docs': doc_requirements
    },
    package_dir={
        '': 'src',
    },
    packages=find_packages(where='src', exclude=('templates',)),
    package_data={'orchd_sdk': [
        'VERSION',
        'logger.ini',
    ]},
    entry_points={
        'console_scripts': [
            'orchd-sdk=orchd_sdk.cli:cli',
        ]
    }
)



#!/usr/bin/env python3

# The MIT License (MIT)
# Copyright © 2022 <Mathias Santos de Brito>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the “Software”), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import importlib
import json
import os
import subprocess
import sys
from time import sleep

import click

from orchd_sdk import version
from orchd_sdk.models import Project
from orchd_sdk.project import ReactionBootstrapper, ProjectBootstrapper, ReactionProperties, SensorProperties, \
    SensorBootstrapper, \
    SinkProperties, SinkBootstrapper
from orchd_sdk.util import is_snake_case, is_kebab_case

PROJECT = None

try:
    with open('orchd.meta.json') as fd:
        PROJECT = Project(**json.load(fd))
except FileNotFoundError as e:
    PROJECT = False
except json.JSONDecodeError:
    click.echo("Project file malformed.")
    sys.exit(-1)


def save_project(project: Project):
    with open(f'orchd.meta.json', 'w') as fd:
        fd.write(json.dumps(project.dict(), indent=2) + '\n\n')


def snake_case_validator(_, __, value):
    if is_snake_case(value):
        return value
    raise click.BadParameter('Name must be snake case. e.g "some_valid_name"')


def kebab_case_validator(_, __, value):
    if is_kebab_case(value):
        return value
    raise click.BadParameter('Name of must be kebab case. e.g "some-valid-name"')


def check_twine_is_present():
    try:
        import twine
    except ImportError:
        click.echo('Twine is required to deploy, and it is not present! '
                   'Install it with "pip install twine" and try again!')
        sys.exit(-1)


def check_current_dir_is_project_root(ctx):
    if not PROJECT and not ctx.invoked_subcommand == 'project':
        print('Error! Be sure to be in a orchd-sdk project root!')
        sys.exit(-1)


def echo_operation_result(operation: str, result: bool):
    click.secho(operation + '... ', bold=True, nl=False)
    if result:
        click.secho('OK!', fg='green')
    else:
        click.secho(' FAILED!', fg='red')
    sleep(1)


def echo_warning(message: str):
    click.secho('WARNING... ', bold=True, nl=False)
    click.secho(message, fg='yellow')
    sleep(1)


def run_shell_command(parameters, cmd_display_name, exit_on_error=True, error_msg='Error. Aborting!'):
    result = subprocess.run(parameters)
    echo_operation_result(cmd_display_name, not bool(result.returncode))
    if result.returncode != 0 and exit_on_error:
        click.echo(error_msg)
        sys.exit(-1)


@click.group()
def actions_group():
    pass


@actions_group.command('template')
@click.option('--from', 'from_', required=True)
def generate_template(from_):
    """Generate a template from an Orchd Asset."""
    split_name = from_.split('.')
    module = '.'.join(split_name[:-1])
    class_name = split_name[-1]
    try:
        Class = getattr(importlib.import_module(module), class_name)
        click.echo(Class.template.json(indent=2))
    except ModuleNotFoundError as e:
        click.echo(f'{e}! Is the related package installed? Try "pip install the_package_containing_module"')
    except AttributeError as e:
        click.echo(f'Module {module} seems not to have a class named {class_name}')


@click.group()
def new_actions_group():
    pass


@new_actions_group.command('reaction')
@click.option('--name', '-n', 'reaction_module_name', callback=snake_case_validator, prompt="Reaction module name (use snake case)")
@click.option('--version', '-v', prompt='Reaction version')
@click.option('--triggers', '-t', prompt='List of event triggers (ex. '
                                         '["io.orchd.events.system.Test", "com.example.events.SomeEvent2"])',
              )
@click.option('--handler_params', '-hp', prompt='Handler parameters (json format)')
@click.option('--active', '-a', default=True, prompt='Active reaction after load?', type=click.types.BOOL)
def new_reaction(reaction_module_name, version, triggers, handler_params, active):
    """Creates a new Reaction"""
    global PROJECT
    handler_params = handler_params or {}
    triggers = triggers or '[]'
    sinks = []

    click.echo('Setting up reaction...')

    reaction_properties = ReactionProperties(PROJECT, reaction_module_name,
                                             version, triggers, sinks,
                                             handler_params, active)
    ReactionBootstrapper.create(PROJECT, reaction_properties)

    PROJECT.add_reaction(reaction_properties.reaction_namespaced_name(),
                         reaction_properties.reaction_class_fq_name())
    save_project(PROJECT)

    click.echo('Done!')


@new_actions_group.command('sensor')
@click.option('--name', '-n', 'sensor_module_name', callback=snake_case_validator,
              help='Sensor module name (use snake case)', prompt='Sensor module name (use snake case)')
@click.option('--description', '-d', default='', help='Brief description for the new sensor.',
              prompt='Brief description of the Sensor', show_default=True)
@click.option('--version', '-v', default='0.0', help='Version number for the sensor.',
              prompt='Sensor version', show_default=True)
@click.option('--sensor-param', '-sp', help='Sensor Parameters as JSON',
              prompt='Sensor parameters as JSON')
@click.option('--sensing-interval', '-si', default=1, help='Sensing Interval in seconds (int)', type=click.INT,
              prompt='Sensing interval', show_default=True)
@click.option('--communicator', '-c', default='orchd_sdk.sensor.LocalCommunicator',
              show_default=True)
def new_sensor(sensor_module_name, description, version, sensor_param, sensing_interval, communicator):
    """Creates a new sensor."""
    click.echo('Setting up new Sensor...')
    sensor_properties = SensorProperties(PROJECT, sensor_module_name,
                                         description, version, sensor_param,
                                         sensing_interval, communicator)
    SensorBootstrapper.create(PROJECT, sensor_properties)

    PROJECT.add_sensor(sensor_properties.sensor_namespaced_name(),
                       sensor_properties.sensor_class_fq_name())
    save_project(PROJECT)
    click.echo('Done!')


@new_actions_group.command('sink')
@click.option('--name', '-n', callback=snake_case_validator, help='Name of the Sink module. (snake case)',
              prompt='Name of the Sink module (snake case)')
@click.option('--version', '-v', help='Sink version', prompt='Sink Version')
@click.option('--parameters', '-p', help='sink parameters', prompt='Sink Parameters (JSON Format)')
def new_sink(name, version, parameters):
    """Creates a new sensor"""
    sink_properties = SinkProperties(PROJECT, name, version, parameters)
    SinkBootstrapper.create(PROJECT, sink_properties)
    PROJECT.add_sink(sink_properties.sink_namespaced_name(), sink_properties.sink_class_name())
    save_project(PROJECT)

    click.echo('SUCCESS!')
    click.echo(f'Sink name is: {sink_properties.sink_namespaced_name()}')
    click.echo(f'Sink module created as: {sink_properties.sink_module_fq_name()}')
    click.echo(f'Sink class reference is: {sink_properties.sink_class_fq_name()}')


@new_actions_group.command('project')
@click.option('--name', '-n', callback=kebab_case_validator, help='Name of the project', prompt='Give a name for the project')
@click.option('--namespace', '-ns', help='Project namespace', prompt='Give the project a namespace (ex. com.example)')
@click.option('--author', '-a', help='Project author', prompt='Project author\'s name')
@click.option('--license', '-l', help='Project license', prompt='Project\'s license')
@click.option('--version', '-v', help='Version of the Project', prompt='Project version')
@click.option('--description', '-d', help='Project description', prompt='Project description')
def new_project(**kwargs):
    """Create a new Project"""
    global PROJECT
    if not PROJECT and not os.path.exists(f'./{kwargs.get("name")}'):
        PROJECT = Project(**kwargs)
        ProjectBootstrapper.setup_project(PROJECT)
    else:
        click.echo('Not able to create the project. A project already exists with this name or in this directory.')


@actions_group.command('test')
def test():
    """Run project tests"""
    run_shell_command(['pytest', '-q'], cmd_display_name='TESTS')


@actions_group.command('build')
@click.option('--skip-tests', help='Skip tests.', is_flag=True)
@click.pass_context
def build(ctx, skip_tests):
    """Build the project and generates the packages"""
    if not skip_tests:
        ctx.invoke(test)

    run_shell_command(['python', 'setup.py', 'sdist', 'bdist_wheel'], cmd_display_name='BUILD',
                      error_msg='Error, Not able to build.')


@actions_group.command('deploy')
@click.option('--upload', help='Deploy definitely the package.', is_flag=True)
@click.option('--skip-tests', help='Skip tests', is_flag=True)
@click.pass_context
def deploy(ctx, upload, skip_tests):
    """Deploy your project to PyPi"""
    check_twine_is_present()
    ctx.invoke(build, skip_tests=skip_tests)
    upload_package_to_pypi(upload)


def upload_package_to_pypi(test_upload=False):
    if test_upload:
        run_shell_command(['twine', 'upload'], cmd_display_name='UPLOAD')
    else:
        upload_package_test()


def upload_package_test():
    click.echo('We will do a test at PyPi Test Server (https://test.pypi.org) '
               'before uploading to PyPi. You need an account there. YOUR PyPi '
               'CREDENTIALS will not work.')
    click.confirm('Are you ready to proceed?', abort=True)
    run_shell_command(['twine', 'upload', '--skip-existing', '-r', 'testpypi', './dist/*'],
                      cmd_display_name='UPLOAD TEST')
    echo_warning('Package was not Upload to production Server yet! (testpypi)!')
    click.echo('Go to test.pypi.org and verify the package.')
    click.secho('To deploy to PyPi run "orchd-sdk deploy --upload"', bold=True)


@actions_group.command(cls=click.CommandCollection, sources=[new_actions_group])
@click.pass_context
def new(ctx):
    """Create new Orchd Assets (Reactions, Sensors, Sinks)"""
    check_current_dir_is_project_root(ctx)


@click.group(cls=click.CommandCollection, sources=[actions_group])
@click.version_option(version())
@click.pass_context
def cli(ctx):
    pass
